(* 
      Certified Grobner bases computations,
      	from the proof of Buchberger algorithm in Coq, 
      	by Laurent Thery.




      All algorithms on polynomials are proven to be correct, except:

      	 - operations on coefficients (which are supposed to be in a field).
         - parsing and pretty-printing of polynomials.

      Coefficients are in the field Z/pZ, with p=31991.
      
          	   Loic Pottier.
      



    The code is seperated in 5 files:

- buch.ml: 
the ocaml code automatically extracted from the Coq specifications and proofs.

- modp.ml:
operations in Z/pZ, with p=31991.

- parsepol.ml:
a parser for polynomials, in Maple-like syntax.

- printpol.ml:
a pretty-printer for polynomials, in Maple-like syntax.

- gbcoq.ml: 
the main program, reading a file, computing grobner basis, and printing it.
The input file contains the list of variables and the list of polynomials. For instance for cyclic5:

[a,b,c,d,e]


[a*b*c*d*e -1,
a*b*c*d + b*c*d*e + c*d*e*a + d*e*a*b + e*a*b*c,
a*b*c + b*c*d + c*d*e + d*e*a + e*a*b,
a*b + b*c + c*d + d*e + e*a,
a+b+c+d+e]

    To compile into native code:

ocamlopt -o gbcoq buch.ml modp.ml printpol.ml parsepol.ml gbcoq.ml -inline 10 

    To run:

time ./gbcoq -file cyclic5

*)

open Buch;;
open Modp;;

(***********************************************************************

      Conversion from int to Peano.
*)

let rec n_to_p n  =  if n = 0 then O else (S (n_to_p (n-1)));;

let eqd n m = if n=m then  Left_ren else Right_ren;;

let plusP =  
  fun n -> 
   let n = n_to_p n in 
   (lplus eqd (ri 0) n (pluscoef) (total_orderc_dec n) n);;

let multP  = 
  fun n -> 
   let n = n_to_p n in 
  (lmult eqd (ri 0) n (pluscoef) (multcoef) (total_orderc_dec n) n);;


let scalP  = 
  fun n -> 
   let n = n_to_p n in 
      fun m -> 
      (lscal eqd (ri 0)  n (multcoef) (ri m));;

let p0 = fun n -> (lpO  eqd (ri 0) (n_to_p n));;

let p1  = (fun n -> (lp1 eqd (ri 0) (n_to_p n) (ri 1))) ;;

let mon = fun n -> fun m -> lgen eqd (ri 0) (n_to_p n) (ri 
1) (n_to_p m);;

let tbuch  =
    (fun n ->
      let n = n_to_p n in 
      tbuch eqd (ri 0) n pluscoef
       oppcoef subcoef  multcoef (total_orderc_dec n) (ri 1) 
divcoef);;

let rec l2l l = match l with [] -> Nil | (a::tl) -> Cons (a, l2l tl);;
let rec l5l l = match l with Nil -> [] | Cons (a,tl) -> a :: (l5l tl);;

let tbuchl = fun n -> fun l -> (l5l (tbuch n (l2l l)));;

(**********************************************************************

      Printing polynomials.
*)

open Format;;
open Printpol;;

let name_var= ref ["X";"Y";"Z"]  (* default *)
;;

let string_of_S a = 
  let rec s_S a d = match a with
     (S b) -> s_S b (1+d)
    | O -> string_of_int d
  in s_S a 0;;

let rec p_to_n  x = match x with
   O -> 0
  | (S n) -> 1+ (p_to_n n)
;;

let printP = print_pol 
    (fun p -> match p with Nil -> true | _ -> false)
    (fun p -> match p with Cons(t,p) -> t)
    (fun p -> match p with Cons(t,p) -> p)
    (fun (Pair (a,m)) -> a)
    (fun (Pair (a,m)) -> m)
    string_of_int
    (fun (C_n (d,_,_)) -> 1+(p_to_n d))
    (fun m i -> let rec gexp m j =
                 match m with
                   (C_n (k,a,b)) -> if j=1 
                                    then a 
                                    else (gexp b (j-1))
                in (string_of_S (gexp m i)))
    name_var
;;

let rec lprintP l =
  match l with
    [] -> ()
   |p::l -> printP p;print_string "\n"; lprintP l
;;

(**********************************************************************

      The main function.

*)

open Parsepol;;

let input = ref "";;

let line = ["-file",Arg.String (fun s -> input:=s),
            "filename"];;

let anon s = raise (Arg.Bad ("What do to with " ^ s ^ "?"));;

Arg.parse line anon "Usage:";;

let varP lv v =
  let ht = Hashtbl.create 10 in
  let dim = List.length lv in
  let rec numvar i l = match l with
     [] -> ()
   | v::l -> Hashtbl.add ht v i; numvar (i+1) l
  in numvar 1 lv;
  let i=Hashtbl.find ht v in
        mon dim (i-1)
;;

let coefP n c =
  (scalP n c (p1 n))
;;

let main() =
  let st =open_in !input in
  let s = ref "" in
  let lv=parse_lvar(input_line st) in
  let rec other() =
    ( try 
	s := !s ^ (input_line st) ;other()
      with End_of_file ->"" ) 
  in other();
  let dim=(List.length lv) in
  let lp=(parse_lpol 
    (p0 dim) (p1 dim) (varP lv) (coefP dim) (plusP dim) (multP dim) !s) in
  name_var:=lv;
  lprintP lp;
  lprintP (tbuchl dim lp)
;;

main();;
