(* 
      Certified Grobner bases computations,
      	from the proof of Buchberger algorithm in Coq, 
      	by Laurent Thery.




      In this version, only the Grobner basis algorithm is proven 
      to be correct: algorithms on polynomials are hand written 
      (addition, division, etc) in a more efficient representation than in 
      the extracted code from Coq. Hence they are not formally proven.
      
      Polynomials are lists of pairs (coefficient, monomial).
      Coefficients are in the field Z/pZ, with p=31991.
      Monomials are arrays of int, total degree in position 0.
      
          	   Loic Pottier.
      



    The code is seperated in 4 files:

- modp.ml:
operations in Z/pZ, with p=31991.

- parsepol.ml:
a parser for polynomials, in Maple-like syntax.

- printpol.ml:
a pretty-printer for polynomials, in Maple-like syntax.

- gbcoq2.ml: 
the main program, reading a file, computing grobner basis, and printing it.
The input file contains the list of variables and the list of polynomials. For instance for cyclic5:

[a,b,c,d,e]


[a*b*c*d*e -1,
a*b*c*d + b*c*d*e + c*d*e*a + d*e*a*b + e*a*b*c,
a*b*c + b*c*d + c*d*e + d*e*a + e*a*b,
a*b + b*c + c*d + d*e + e*a,
a+b+c+d+e]

    To compile into native code:

ocamlopt -o gbcoq2 modp.ml printpol.ml parsepol.ml gbcoq2.ml -inline 10 

    To run:

time ./gbcoq2 -file cyclic5

*)

(***********************************************************************

    Monomials.

*)

let hmon = Hashtbl.create 1000;;

type mon = int array;;

let mult_mon d m m' =
  let m'' = Array.create (d+1) 0 in
  for i=0 to d do
     m''.(i)<- (m.(i)+m'.(i));
  done;
  m''
;;

let deg m = m.(0)
;;

(* lexinv 
let m' = [|6;1;2;4|];;
let m = [|6;1;2;3|];;
*)

let compare_mon d m m' =
  match compare (deg m) (deg m') with
   0 -> let res=ref 0 in
        let i=ref d in
        while (!res=0) && (!i<>0) do
            res:= (compare m.(!i) m'.(!i));
            i:=!i-1;
        done;
        -(!res)
  | x -> x
;;


let div_mon d m m' =
  let m'' = Array.create (d+1) 0 in
  for i=0 to d do
     m''.(i)<- (m.(i)-m'.(i));
  done;
  m''
;;

(* m' divides m *)

let div_mon_test d m m' =
        let res=ref true in
        let i=ref 1 in
        while (!res) && (!i<=d) do
            res:= (m.(!i) >= m'.(!i));
            i:=succ !i;
        done;
        !res
;;

(* moins rapide...
let div_mon_test d m m' =
   let rec divt i= match i with
       0 -> true
      |i -> match (m.(i) >= m'.(i)) with
             true -> divt (i-1)
            |false -> false
   in divt d
;;
*)     
let set_deg d m =
  m.(0)<-0;
  for i=1 to d do
     m.(0)<-  m.(i)+m.(0);
  done;
  m
;;

let ppcm_mon d m m' =
  let m'' = Array.create (d+1) 0 in
  for i=1 to d do
     m''.(i)<- (max m.(i) m'.(i));
  done;
  set_deg d m''
;;

open Modp;;

(**********************************************************************

    Terms and polynomials.

*)

type term = int * mon;;

type pol = term list;;

let plusP d p q =
 let rec plusP p q =
  match p with
   [] -> q
  |t::p' -> 
     match q with
       [] -> p
      |t'::q' ->
          match compare_mon d (snd t) (snd t') with
            1 -> t::(plusP p' q)
           |(-1) -> t'::(plusP p q')
           |_ -> let c=pluscoef (fst t) (fst t') in
                 match eq_coef c coef0 with
                   true -> (plusP p' q')
                  |false -> (c,(snd t))::(plusP p' q')
 in plusP p q
;;

let mult_t_pol d a m p =
  let rec mult_t_pol p =
    match p with
      [] -> []
     |(b,m')::p -> ((multcoef a b),(mult_mon d m m'))::(mult_t_pol p)
  in mult_t_pol p
;;

(* rend un polynome de l dont le monome de tete divise m,  [] si rien *)

let rec selectdiv d m l =
  match l with
    [] -> []
   |q::l -> let ((_,m')::_) = q in
            match (div_mon_test d m m') with
               true -> q
              |false -> selectdiv d m l
;;


let gen d i =
  let m = Array.create (d+1) 0 in
  m.(i) <- 1;
  set_deg d m;
  [((ri 1),m)]
;;

let emultP d a p =
  let rec emultP p =
    match p with
      [] -> []
     |(b,m')::p -> ((multcoef a b),m')::(emultP p)
  in emultP p
;;

let normalize d p =
  match p with
   [] -> []
  |(a,m)::p' -> (emultP d (invmod a) p)
;;

(* divise p par q: rend le reste.
on a lt(p) = -a*m*lt(q)
*)

let ndiv = ref 0;;

let infodiv s =
    (* print_string s;flush stdout; *)
    ndiv:=!ndiv+1
;;

let div_pol d p q a m = infodiv ".";
                         plusP d p (mult_t_pol d a m q)
;;

(* reste de la division de p par les polynomes de l *)

let reduce2 d p l =
  let rec reduce p =
   match p with
    [] -> []
   |t::p' -> let (a,m)=t in
             let q = (try Hashtbl.find hmon m 
                            with Not_found -> 
                             let q = selectdiv d m l in
                             match q with 
                               t'::q' -> (Hashtbl.add hmon m q;q)
                              |[] -> q) in
             match q with
                  [] -> t::(reduce p')
                 |(b,m')::q' -> (reduce (div_pol d p' q'
		                         (oppcoef
					   (divcoef a b))
                                         (div_mon d m m')))
  in reduce p
;;

let reduce d p l =
  let rec reduce p =
   match p with
    [] -> []
   |t::p' -> let (a,m)=t in
             let q = selectdiv d m l in
             match q with
                  [] -> t::(reduce p')
                 |(b,m')::q' -> (reduce (div_pol d p' q'
		                         (oppcoef
					   (divcoef a b))
                                         (div_mon d m m')))
  in reduce p
;;

(* m'' = ppcm des monomes de tete de p et q *)

let spol d p q=
   let (a,m)::p' = p in
   let (b,m')::q' = q in
   let m''=(ppcm_mon d m m') in
   plusP d (mult_t_pol d (invmod a) (div_mon d m'' m) p')
           (mult_t_pol d (oppcoef (invmod b)) (div_mon d m'' m') q')
;;

(* spolynomes provoques par l'ajout d'un polynome *)

type 'poly cpRes =
    Keep of ('poly list)
  | DontKeep of ('poly list)
;;



let foreigner d p p'=
        let (_,m)::_=p in
        let (_,m')::_=p' in
        let res=ref true in
        let i=ref 1 in
        while (!res) && (!i<=d) do
            res:= (m.(!i) = 0) || (m'.(!i)=0);
            i:=!i+1;
        done;
        !res
;;

let list_rec f0 f1 =
  let rec f2 = function
    [] -> f0
  | a0::l0 -> f1 a0 l0 (f2 l0)
  in f2
;;

let div_ppcm d p p' p'' =
        let (_,m)::_=p in
        let (_,m')::_=p' in
        let (_,m'')::_=p'' in
        let res=ref true in
        let i=ref 1 in
        while (!res) && (!i<=d) do
            res:= ((max m.(!i) m'.(!i)) >= m''.(!i));
            i:=!i+1;
        done;
        !res
;;

let addRes i = function
    Keep h'0 -> Keep (i::h'0)
  | DontKeep h'0 -> DontKeep (i::h'0)
;;


let npairelim = ref 0;;

let infoelim s =
    (* print_string s;flush stdout;*)
    npairelim:=!npairelim+1
;;

let slice d i a q =
  list_rec
    (match foreigner d i a with
       true -> infoelim "*";DontKeep []
     | false -> Keep [])
  (fun b q1 rec_ren ->
    match div_ppcm d i a b with
      true ->  infoelim "+";DontKeep (b::q1)
    | false ->
        (match div_ppcm d i b a with
           true -> infoelim "#";rec_ren
         | false -> addRes b rec_ren)) q
;;

let slice2 d i a q =
   let rec slice1 q =
     match q with
       [] ->  []
      |b::q1 -> match div_ppcm d i b a with
                          true -> infoelim "#"; slice1 q1
                        | false -> b::(slice1 q1)
   in let rec slice2 q =
        match q with
          [] ->  Keep []
         |b::q1 ->
            match div_ppcm d i a b with
              true -> infoelim "+"; DontKeep (b::(slice1 q1))
            | false -> (match div_ppcm d i b a with
                          true -> infoelim "#"; slice2 q1
                        | false -> addRes b (slice2 q1))
   in match foreigner d i a with
       true -> infoelim "*"; DontKeep (slice1 q)
     | false -> (slice2 q)
;;
(* 
slice:
Paires eliminees: 2136
Divisions: 1648
Paires reduites a zero: 75
Paires donnant un nouveau polynome: 37
temps natif: 3.54s user

slice2:
Paires eliminees: 745
Divisions: 1724
Paires reduites a zero: 79
Paires donnant un nouveau polynome: 37
temps natif: 3.51s user
*)
let rec addS x l = l @[x]
;;

let genPcPf d i aP q =
  (let rec genPc aP0 =
    match aP0 with
      [] -> (fun r -> r)
    | a::l1 -> (fun l ->
        (match slice d i a l1 with
          Keep l2 -> addS (spol d i a) (genPc l2 l)
        | DontKeep l2 -> genPc l2 l))
 in genPc aP) q
;;

let genOCPf d h' =
  list_rec [] (fun a l rec_ren ->
    genPcPf d a l rec_ren) h'
;;

let rec redd d l l2 = match l with 
|  [] -> []
| (a::p) -> let l22= l2@p in
                  let pol = normalize d (reduce d a l22) in
                 ( match pol with 
                  [] -> redd d p l2 
                 | _ -> (pol::(redd d p (pol::l2))))
;;


(*
     Printing polynomials.
*)

open Printpol;;

let name_var= ref []
;;

let printP = print_pol 
    (fun p -> match p with [] -> true | _ -> false)
    (fun p -> match p with (t::p) -> t)
    (fun p -> match p with (t::p) -> p)
    (fun (a,m) -> a)
    (fun (a,m) -> m)
    string_of_int
    (fun m -> (Array.length m)-1)
    (fun m i -> (string_of_int (m.(i))))
    name_var
;;

let rec lprintP l =
  match l with
    [] -> ()
   |p::l -> printP p;print_string "\n"; lprintP l
;;

(* #install_printer printP;; *)

(* buchberger *)


let prod_rec f = function
  (x0, x) -> f x0 x
;;

let npairzero = ref 0;;

let infonpairzero s =
    print_string s;flush stdout;
    npairzero:=!npairzero+1
;;

let npairnew = ref 0;;

let infonpairnew s =
    print_string s;flush stdout;
    npairnew:=!npairnew+1
;;

let step = ref 0;;

let infobuch p q =
   if !step = 0 
      then (print_string ("[" ^ (string_of_int (List.length p))
                          ^ "," ^  (string_of_int (List.length q))
                          ^ "]");flush stdout)
;;
(***********************************************************************)
let pbuchf d pq =
  step:=0;
  Hashtbl.clear hmon;
  let rec pbuchf x = 
    prod_rec (fun p q ->
      infobuch p q;
      step:=(!step+1)mod 10;
      match q with
        [] -> p
      | (a::q2) -> 
          let a0=normalize d (reduce2 d a p) in
            match a0 with
              [] -> infonpairzero "0";pbuchf (p, q2)
            | _ ->(* printP a0;*)
                infonpairnew "n";
                 pbuchf
                  (((addS a0 p),
                         (genPcPf d a0 p q2)))) x
  in pbuchf pq 
;;

let buch d p =
  npairelim:= 0;
  ndiv:= 0;
  npairzero:= 0;
  npairnew:= 0;
  let r = redd d (pbuchf d (p, (genOCPf d p))) [] in
  print_string "\nPaires eliminees: ";print_int !npairelim;
  print_string "\nDivisions: ";print_int !ndiv;
  print_string "\nPaires reduites a zero: ";print_int !npairzero;
  print_string "\nPaires donnant un nouveau polynome: ";print_int !npairnew;
  print_string "\n";
  r     
;;

let rec multP d p q =
  match p with 
    [] -> []
   |(a,m)::p' -> (plusP d (mult_t_pol d a m q) (multP d p' q))
;;

(**********************************************************************

      The main function.

*)

open Parsepol;;

let input = ref "";;

let line = ["-file",Arg.String (fun s -> input:=s),
            "filename"];;

let anon s = raise (Arg.Bad ("What do to with " ^ s ^ "?"));;

Arg.parse line anon "Usage:";;

let p0 n = [];;
let p1 n = [(1,Array.create (n+1) 0)];;
let varP lv v =
  let ht = Hashtbl.create 10 in
  let dim = List.length lv in
  let rec numvar i l = match l with
     [] -> ()
   | v::l -> Hashtbl.add ht v i; numvar (i+1) l
  in numvar 1 lv;
  let m=Array.create (dim+1) 0 in
     m.(Hashtbl.find ht v) <- 1;
     m.(0) <- 1;
     [(1,m)]
;;

let coefP n c =
  let m=Array.create (n+1) 0 in [(c,m)]
;;

let main() =
  let st =open_in !input in
  let s = ref "" in
  let lv=parse_lvar(input_line st) in
  let rec other() =
    ( try 
	s := !s ^ (input_line st) ;other()
      with End_of_file ->"" ) 
  in other();
  let dim=(List.length lv) in
  let lp=(parse_lpol 
    (p0 dim) (p1 dim) (varP lv) (coefP dim) (plusP dim) (multP dim) !s) in
  name_var:=lv;
  lprintP lp;
  lprintP (buch dim lp)
;;

main();;

